import {AnyThreadChannel, ForumChannel, Message} from 'discord.js';
import {writeFile} from 'fs/promises';
import {BotContext} from '../common';
import {
	ensureChannelContent,
	getChannelIdPath,
	getMetadata,
	setMetadata,
} from './channel';

const getChannelThreadIdPath = (channel: ForumChannel, thread: AnyThreadChannel) => `${getChannelIdPath(channel)}/${thread.id}`;

const serializeMessage = (message: Message) => {
	const messageContent = [
		message.content,
		...Array.from(message.attachments.values()).map((a) => (
			a.contentType?.startsWith('image/') ? `[![${a.name}](${a.url})](${a.url})` : `[${a.name}](${a.url})`
		))
	];
	return (
		`\n${message.id} ${message.createdTimestamp} ${message.author.id} ${JSON.stringify(messageContent.join(' ').trim())}`
	);
}

const serializeThreadMetadata = (thread: AnyThreadChannel) => (
	`${thread.id} ${thread.createdTimestamp} ${thread.client.user?.id} ${JSON.stringify(thread.name)}`
);

export const addThreadMessage = async (channel: ForumChannel, thread: AnyThreadChannel, message: Message) => {
	try {
		await writeFile(
			getChannelThreadIdPath(channel, thread),
			serializeMessage(message),
			{ flag: 'a+' },
		);
	} catch {
		// noop
	}
};

export const addThread = async (channel: ForumChannel, thread: AnyThreadChannel) => {
	await ensureChannelContent(channel);

	try {
		await writeFile(
			getChannelThreadIdPath(channel, thread),
			serializeThreadMetadata(thread),
		);
	} catch {
		// noop
	}

	try {
		// todo how to fetch unrecorded messages
		const messages = await thread.messages.fetch();
		const collection = messages.sorted((a, b) => a.createdTimestamp - b.createdTimestamp);
		const collectionArray = Array.from(collection.values());
		await collectionArray.reduce(
			async (promise, message) => {
				await promise;
				return addThreadMessage(channel, thread, message);
			},
			Promise.resolve(),
		);
	} catch {
		// noop
	}
};

export const setupThreads = async (context: BotContext) => {
	if (typeof context.channel === 'undefined') {
		return;
	}
	const channel = context.channel;
	const metadata = await getMetadata(channel);
	// todo how to fetch unrecorded threads
	const threads = await channel.threads.fetch();
	for (const [, thread] of threads.threads) {
		if (typeof metadata.lastSyncedAt !== 'undefined' && (thread.createdTimestamp ?? Infinity) <= metadata.lastSyncedAt) {
			continue;
		}
		await addThread(channel, thread);
	}
	await setMetadata(channel);
};
